import networkx as nx
import plotly.graph_objs as go


class NetworkPlotly:

    def __init__(self, name='Zettelkasten'):
        """
        Build network to visualize with Plotly

        Parameters
        ----------
        name : str
            The network name.
        """
        self.graph = nx.Graph()

    def add_node(self, node_id, title):
        """
        Add a node to the network.

        Parameters
        ----------
        node_id : str, or int
            A  unique identifier for the node, typically the zettel ID.
        title : str
            The text label for each node, typically the zettel title.

        """
        self.graph.add_node(node_id, title=title)

    def add_edge(self, source, target):
        """
        Add a node (a zettel) to the network.

        Parameters
        ----------
        source : str or int
            The ID of the source zettel.
        target : str or int
            The ID of the target (cited) zettel.

        """
        self.graph.add_edge(source, target)

    def build_plotly_figure(self, pos=None):
        """
        Creates a Plot.ly Figure that can be view online or offline.

        Parameters
        ----------
        graph : nx.Graph
            The network of zettels to visualize
        pos : dict
            Dictionay of zettel_id : (x, y) coordinates where to draw nodes. If
            None, the Kamada Kawai layout will be used.

        Returns
        -------
        fig : plotly Figure

        """

        if pos is None:
            # The kamada kawai layout produces a really nice graph but it's
            # a O(N^2) algorithm. It seems only reasonable to draw the graph
            # with fewer than ~1000 nodes.
            if len(self.graph) < 1000:
                pos = nx.layout.kamada_kawai_layout(self.graph)
            else:
                pos = nx.layout.random_layout(self.graph)

        # Create scatter plot of the position of all notes
        node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode='markers',
            hoverinfo='text',
            marker=dict(
                showscale=True,
                # colorscale options
                #'Greys' | 'YlGnBu' | 'Greens' | 'YlOrRd' | 'Bluered' | 'RdBu' |
                #'Reds' | 'Blues' | 'Picnic' | 'Rainbow' | 'Portland' | 'Jet' |
                #'Hot' | 'Blackbody' | 'Earth' | 'Electric' | 'Viridis' |
                colorscale='YlGnBu',
                reversescale=True,
                color=[],
                size=10,
                colorbar=dict(
                    thickness=15,
                    title='Centrality',
                    xanchor='left',
                    titleside='right'
                ),
                line=dict(width=2)))

        for node in self.graph.nodes():
            x, y = pos[node]
            text = '<br>'.join([node, self.graph.node[node].get('title', '')])
            node_trace['x'] += tuple([x])
            node_trace['y'] += tuple([y])
            node_trace['text'] += tuple([text])

        # Color nodes based on the centrality
        for node, centrality in nx.degree_centrality(self.graph).items():
            node_trace['marker']['color']+=tuple([centrality])

        # Draw the edges as annotations because it's only sane way to draw arrows.
        edges = []
        for from_node, to_node in self.graph.edges():
            edges.append(
                dict(
                    # Tail coordinates
                    ax=pos[from_node][0], ay=pos[from_node][1], axref='x', ayref='y',
                    # Head coordinates
                    x=pos[to_node][0], y=pos[to_node][1], xref='x', yref='y',
                    # Aesthetics
                    arrowwidth=2, arrowcolor='#666', arrowhead=2,
                    # Have the head stop short 5 px for the center point,
                    # i.e., depends on the node marker size.
                    standoff=5,
                    )
                )

        fig = go.Figure(
            data=[node_trace],
            layout=go.Layout(
                showlegend=False,
                hovermode='closest',
                margin=dict(b=20, l=5, r=5, t=40),
                annotations=edges,
                xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
                yaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
            )
        )
        return fig

    def render(self, output, view=True):
        """
        Render the network to disk.

        Parameters
        ----------
        output : str
            Name of the output file.
        view : bool
            Open the rendered network using the default browser. Default is
            True.

        """
        fig = self.build_plotly_figure()
        if not output.endswith('.html'):
            output += '.html'
        fig.write_html(output, auto_open=view)



from textwrap import fill

from graphviz import Digraph


class NetworkGraphviz:

    def __init__(self, name='Zettelkasten', engine='sfdp', shape='record'):
        """
        Build network to visualize with Graphviz.

        Parameters
        ----------
        name : str (optional)
            Name of the network. Default is "Zettelkasten".
        engine : str
            Layout engine used by Graphviz. The default is 'sfdp', which is a
            good default. See the graphviz documentation for alternatives.
        shape : str {record, plaintext}
            The shape, or style, of each node. The default is "record".
        """
        self.name = name
        self.engine = engine
        self.graph = Digraph(comment=self.name, engine=self.engine)
        self.shape = shape

    def wrap_title(self, text, width=30):
        """
        Wrap the title to be a certain width.

        Parameters
        ----------
        text : str
            The text to wrap.
        width : int
            The text width, in characters.

        """
        return fill(text, width)

    def add_node(self, node_id, title):
        """
        Add a node to the network.

        Parameters
        ----------
        node_id : str, or int
            A  unique identifier for the node, typically the zettel ID.
        title : str
            The text label for each node, typically the zettel title.

        """
        if self.shape == 'plaintext':
            label = self.wrap_title("{} {}".format(node_id, title)).strip()
        elif self.shape == 'record':
            # Wrap in {} so the elements are stacked vertically
            label = "{" + '|'.join([node_id, self.wrap_title(title)]) + "}"

        self.graph.node(node_id, label, shape=self.shape)

    def add_edge(self, source, target):
        """
        Add a node (a zettel) to the network.

        Parameters
        ----------
        source : str or int
            The ID of the source zettel.
        target : str or int
            The ID of the target (cited) zettel.

        """
        self.graph.edge(source, target)

    def render(self, output, view=True):
        """
        Render the network to disk.

        Parameters
        ----------
        output : str
            Name of the output file.
        view : bool
            Show the network using the default PDF viewer. Default is True.

        """
        self.graph.render(output, view=view)



from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import TestCase

from zkviz import zkviz


class TestListZettels(TestCase):

    def test_list_zettels_with_md_extension(self):
        # Create a temporary folder and write files in it
        with TemporaryDirectory() as tmpdirname:
            ext = '.md'
            basename = '201906242157'
            filepaths = []
            for i in range(3):
                path = Path(tmpdirname, basename + str(i) + ext)
                path.touch()
                filepaths.append(str(path))

            files_found = zkviz.list_zettels(tmpdirname)
            self.assertEqual(filepaths, files_found)

    def test_list_zettels_with_txt_extension(self):
        # Create a temporary folder and write files in it
        with TemporaryDirectory() as tmpdirname:
            ext = '.txt'
            basename = '201906242157'
            filepaths = []
            for i in range(3):
                path = Path(tmpdirname, basename + str(i) + ext)
                path.touch()
                filepaths.append(str(path))

            files_found = zkviz.list_zettels(tmpdirname, '*.txt')
            self.assertEqual(filepaths, files_found)

    def test_list_zettels_with_mixed_extensions(self):
        # Create a temporary folder and write files in it
        with TemporaryDirectory() as tmpdirname:
            filepaths = []
            basename = '201906242157'

            ext = '.txt'
            for i in range(5):
                path = Path(tmpdirname, basename + str(i) + ext)
                path.touch()
                filepaths.append(str(path))

            ext = '.md'
            for i in range(5, 10):
                path = Path(tmpdirname, basename + str(i) + ext)
                path.touch()
                filepaths.append(str(path))

            files_found = zkviz.list_zettels(tmpdirname, '*.txt|*.md')
            self.assertEqual(filepaths, files_found)


class TestParseArgs(TestCase):

    def test_default_extension(self):
        args = zkviz.parse_args('')
        self.assertEqual(['*.md'], args.pattern)

    def test_overwrite_extension(self):
        args = zkviz.parse_args(["--pattern","*.txt"])
        self.assertEqual(['*.txt'], args.pattern)

    def test_multiple_extensions(self):
        args = zkviz.parse_args(["--pattern","*.txt", "--pattern", '*.md'])
        self.assertEqual(['*.txt', '*.md'], args.pattern)



"""

Visualize the notes network of a Zettelkasten.

Each arrow represents a link from one zettel to another. The script assumes
that zettels have filenames of the form "YYYYMMDDHHMM This is a title" and that
links have the form [[YYYYMMDDHHMM]]

"""
import glob
import os.path
import re
from textwrap import fill




PAT_ZK_ID = re.compile(r'^(?P<id>\d+)\s(.*)')
PAT_LINK = re.compile(r'\[\[(\d+)\]\]')


def parse_zettels(filepaths):
    """ Parse the ID and title from the filename.

    Assumes that the filename has the format "YYYYMMDDHHMMSS This is title"

    """
    documents = []
    for filepath in filepaths:
        basename = os.path.basename(filepath)
        filename, ext = os.path.splitext(basename)
        r = PAT_ZK_ID.match(filename)
        if not r:
            continue

        with open(filepath, encoding='utf-8') as f:
            links = PAT_LINK.findall(f.read())

        document = dict(id=r.group(1), title=r.group(2), links=links)
        documents.append(document)
    return documents


def create_graph(zettels, graph):
    """
    Create of graph of the zettels linking to each other.

    Parameters
    ----------
    zettels : list of dictionaries

    Returns
    -------
    graph : nx.Graph

    """

    for doc in zettels:
        graph.add_node(doc['id'], title=doc['title'])
        for link in doc['links']:
            graph.add_edge(doc['id'], link)
    return graph


def list_zettels(notes_dir, pattern='*.md'):
    """
    List zettels in a directory.

    Parameters
    ----------
    notes_dir : str
        Path to the directory containing the zettels.
    pattern : str (optional)
        Pattern matching zettels. The default is '*.md'. If there are multiple
        patterns, separate them with a |, such as in '*.md|*.txt'

    """

    filepaths = []

    for patt in pattern.split('|'):
        filepaths.extend(glob.glob(os.path.join(notes_dir, patt)))
    return sorted(filepaths)


def parse_args(args=None):
    from argparse import ArgumentParser
    parser = ArgumentParser(description=__doc__)
    parser.add_argument('--notes-dir', default='.',
                        help='path to folder containin notes. [.]')
    parser.add_argument('--output', default='zettel-network',
                        help='name of output file. [zettel-network]')
    parser.add_argument('--pattern', action='append',
            help=('pattern to match notes. You can repeat this argument to'
            ' match multiple file types. [*.md]'))
    parser.add_argument('--use-graphviz', action='store_true', default=False,
            help='Use Graphviz instead of plotly to render the network.')
    parser.add_argument('zettel_paths', nargs='*', help='zettel file paths.')
    args = parser.parse_args(args=args)

    # Use the list of files the user specify, otherwise, fall back to
    # listing a directory.
    if not args.zettel_paths:
        if args.pattern is None:
            args.pattern = ['*.md']
        patterns = '|'.join(args.pattern)

        args.zettel_paths = list_zettels(args.notes_dir, pattern=patterns)
    return args


def main(args=None):
    args = parse_args(args)

    zettels = parse_zettels(args.zettel_paths)

    # Fail in case we didn't find a zettel
    if not zettels:
        raise FileNotFoundError("I'm sorry, I couldn't find any files.")

    if args.use_graphviz:
        from zkviz.graphviz import NetworkGraphviz
        import graphviz
        try:
            graphviz.version()
        except graphviz.ExecutableNotFound:
            raise FileNotFoundError(fill(
                "The Graphviz application must be installed for the"
                " --use-graphviz option to work. Please see"
                " https://graphviz.org/download/ for installation"
                " instructions."
           ))
        graph = NetworkGraphviz()
    else:
        from zkviz.plotly import NetworkPlotly
        graph = NetworkPlotly()

    graph = create_graph(zettels, graph)
    graph.render(args.output)


if __name__ == "__main__":
    import sys
    try:
        sys.exit(main())
    except FileNotFoundError as e:
        # Failed either because it didn't find any files or because Graphviz
        # wasn't installed
        sys.exit(e)



import os.path
import re
import setuptools


# Find the version number
here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with open(os.path.join(here, *parts), 'r', encoding='utf-8') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# Load README
with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="zkviz",
    version=find_version("zkviz", "__init__.py"),
    author="Alexandre Chabot-Leclerc",
    author_email="github@alexchabot.net",
    description="Zettel Network Visualizer",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Zettelkasten-Method/zkviz",
    packages=setuptools.find_packages(),
    install_requires=[
        'graphviz',
        'plotly',
        'numpy',
        'scipy',
        'networkx',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            'zkviz = zkviz.zkviz:main'
        ]
    }
)
